function overrideMethod(target, methodName, from, newMethodBuilder) {
    const temp0 = `original__${methodName}__${from}__is_overriden`;
    if (target[temp0] === true) {
        return; // Do not override twice;
    }
    const temp = `original__${methodName}__${from}`;
    if (!target[temp] && typeof target[methodName] === 'function') {
        target[temp] = target[methodName].bind(target);
        target[temp].bind(target);
    }
    const alreadyAttached = target[temp];
    if (alreadyAttached) {
        const altered = newMethodBuilder(alreadyAttached);
        const bind = altered.bind(target);
        target[methodName] = bind;
    }
    else {
        const altered = newMethodBuilder(null);
        const bind = altered.bind(target);
        target[methodName] = bind;
    }
    target[temp0] = true;
}
function* generateId() {
    let base = 0;
    while (true) {
        base += 1;
        yield base;
    }
}
/**
 * Note: Stencil js has an async `forceUpdate` that batches the rendering.
 *       It is safe to use this class with 'list' and 'plain' in a Stencil context;
 */
class RenderQ {
    constructor(structure, method, forceUpdate) {
        this.list = [];
        this.set = new Set();
        this.structure = structure;
        this.method = method;
        this.forceUpdate = forceUpdate;
    }
    add(component) {
        if (this.structure === 'list') {
            this.list.push(component);
        }
        else {
            this.set.add(component);
        }
    }
    values() {
        if (this.structure === 'list') {
            return this.list;
        }
        else {
            return this.set.values();
        }
    }
    clear() {
        if (this.structure === 'list') {
            this.list = [];
        }
        else {
            this.set.clear();
        }
    }
    signalToRender() {
        const values = this.values();
        if (this.method === 'plain') {
            this.signalToRenderPlain(values);
        }
        else {
            this.signalToRenderAsync(values);
        }
        this.clear();
    }
    signalToRenderPlain(values) {
        for (const qi of values) {
            this.forceUpdate(qi);
        }
    }
    signalToRenderAsync(values) {
        const vs = [...values];
        new Promise(() => {
            this.signalToRenderPlain(vs);
        });
    }
}
function createStore(initialState, rootReducer, config) {
    const { forceUpdate, getElement, getRenderingRef } = config;
    let subs = [];
    let stash = Object.assign({}, initialState);
    const reducer = rootReducer;
    const getState = () => stash;
    const renderQ = new RenderQ('set', 'async', forceUpdate);
    const handleAction = (action) => {
        const crtStash = stash;
        const newStash = reducer(crtStash, action);
        subs = subs.filter((sub) => sub.caller2.isConnected);
        for (const sc of subs) {
            const newSelection = sc.sel2(newStash, sc.vars2);
            const oldSelection = sc.sel2(crtStash, sc.vars2);
            if (newSelection !== oldSelection) {
                renderQ.add(sc.caller2);
            }
        }
        stash = newStash;
        // @ts-ignore
        window.stash = stash;
        renderQ.signalToRender();
    };
    function createProvider(_component) {
        // @ts-ignore
        const stash = rootReducer(initialState, { type: '__STASH_INIT__' });
    }
    function createDispatch(component) {
        // const el = getElement(component);
        if (!component['onCleanups']) {
            const onCleaups = [];
            component['onCleanups'] = onCleaups;
        }
        const dispatch = function dispatch(action) {
            handleAction(action);
        };
        const thunkDispatch = function thunkDispatch(action) {
            let thunkOnCleanup = undefined;
            const context = {
                setOnCleanup: (onCleaup) => {
                    thunkOnCleanup = onCleaup;
                    component['onCleanups'].push(onCleaup);
                },
            };
            if (action[Symbol.toStringTag] === 'AsyncFunction') {
                // it is an async thunk
                action(dispatch, getState, context).then((r) => {
                    component['onCleanups'] = component['onCleanups'].filter((oc) => oc !== thunkOnCleanup);
                    return r;
                });
            }
            else if (Object.prototype.toString.call(action) === '[object Function]') {
                // it is a thunk
                action(dispatch, getState, context);
                component['onCleanups'] = component['onCleanups'].filter((oc) => oc !== thunkOnCleanup);
            }
            else {
                // @ts-ignore
                dispatch(action);
            }
        };
        overrideMethod(component, 'disconnectedCallback', 'createDispatch', function (original) {
            return function () {
                for (const onCleanup of component['onCleanups']) {
                    onCleanup === null || onCleanup === void 0 ? void 0 : onCleanup();
                }
                original === null || original === void 0 ? void 0 : original();
            };
        });
        return thunkDispatch;
    } // createDispatch
    const componentIdGenerator = generateId();
    function createSelector(selector, component) {
        const el = getElement(component);
        const componentName = component.constructor.name;
        if (!component['ownsubs']) {
            // component['ownsubs'] = new Set<SEV<S, any, any>>();
            component['ownsubs'] = new Set();
        }
        if (!el['componentId']) {
            const id = componentIdGenerator.next().value;
            el['componentId'] = `${componentName}::${id}`;
        }
        if (!component['idGenerator']) {
            const idGenerator = generateId();
            component['idGenerator'] = idGenerator;
        }
        const idGenerator = component['idGenerator'];
        const selId = idGenerator.next().value;
        const componentId = el['componentId'];
        selector['stashId'] = `${componentId}::${selId}`;
        const select = function select() {
            const rref = getRenderingRef();
            const existingSub = subs.find((sb) => sb.sel2['stashId'] === selector['stashId']);
            // register only if it's called from render
            if (!existingSub && rref) {
                const sub = { sel2: selector, caller2: el, vars2: undefined };
                subs.push(sub);
                component['ownsubs'].add(selector['stashId']);
            }
            const v = selector(stash);
            return v;
        };
        overrideMethod(component, 'disconnectedCallback', 'createSelector', function (original) {
            return function () {
                // subs = subs.filter(sb => sb.sel2['stashId'] === selector['stashId'] && sb.);
                // remove all own subs from main subs
                subs = subs.filter((sb) => !component['ownsubs'].has(sb.sel2['stashId']));
                delete component['ownsubs'];
                original === null || original === void 0 ? void 0 : original();
            };
        });
        function selectorFunc() {
            return select();
        }
        selectorFunc.bind(component);
        return selectorFunc;
    }
    return { createDispatch, createSelector, createProvider };
}
export { createStore };
